﻿using Autofac;
using FitConnect.BaseClasses;
using FitConnect.Encryption;
using FitConnect.Models;
using FitConnect.Models.Api.Metadata;
using FitConnect.RestService;
using FitConnect.Services;
using FitConnect.Services.Models;
using FitConnect.Services.Models.v1;
using Jose;
using Microsoft.Extensions.Logging;
using Newtonsoft.Json;
using Attachment = FitConnect.Models.Attachment;
using Data = FitConnect.Models.Data;
using Metadata = FitConnect.Models.Metadata;

namespace FitConnect;

public class Sender : FunctionalBaseClass {
    public FitEncryption? Encryption { get; protected set; }
    public bool IsAuthenticated => _oAuthToken is { IsValid: true };
    private OAuthAccessToken? _oAuthToken;


    internal Sender(FitConnectEndpoints endpoints,
        IOAuthService oAuthService,
        ISubmissionService submissionService,
        IRouteService routeService,
        IDestinationService destinationService,
        ILogger? logger = null
    ) : base(endpoints,
        logger, oAuthService, submissionService, routeService, destinationService) {
    }

    public static Sender Create(FitConnectEndpoints endpoints,
        ILogger? logger = null) {
        var container = DiContainer.GetContainer(endpoints, logger);
        var sender = container.Resolve<Sender>();
        return sender;
    }

    /// <summary>
    ///     Check public keys
    /// </summary>
    /// <returns></returns>
    public async Task<bool> CheckPublicKeyAsync(string publicKey) {
        throw new NotImplementedException();
    }

    public async Task<string> Encrypt(string publicKey, byte[] attachment) {
        throw new NotImplementedException();
    }


    /// <summary>
    ///     Check public keys
    /// </summary>
    /// <returns></returns>
    public async Task<bool> CheckPublicKeyAsync() {
        throw new NotImplementedException();
    }

    /// <summary>
    ///     Encrypt Data
    /// </summary>
    /// <param name="data"></param>
    /// <returns></returns>
    public string Encrypt(string publicKey, string? data) {
        if (data == null) return "";
        throw new NotImplementedException();
    }

    /// <summary>
    ///     Encrypt attachments (Anhänge)
    /// </summary>
    /// <param name="publicKey"></param>
    /// <param name="attachment"></param>
    /// <returns></returns>
    public KeyValuePair<string, string> Encrypt(string publicKey, Attachment attachment) {
        var content = Encryption.Encrypt(attachment.Content, publicKey);
        return new KeyValuePair<string, string>(attachment.Id, content);
    }

    /// <summary>
    ///     Encrypt attachments (Anhänge)
    /// </summary>
    /// <param name="publicKey">Public key for encryption</param>
    /// <param name="attachments">List of attachments to encrypt</param>
    /// <returns></returns>
    public Dictionary<string, string> Encrypt(string publicKey,
        IEnumerable<Attachment> attachments) {
        return attachments.Select(a => Encrypt(publicKey, a))
            .ToDictionary(p => p.Key, p => p.Value);
    }

    /// <summary>
    ///     Create Metadata incl. Hash
    /// </summary>
    /// <param name="submission"></param>
    /// <returns></returns>
    public string CreateMetadata(Submission submission) {
        var data = new FitConnect.Models.Api.Metadata.Data {
            SubmissionSchema = new Fachdatenschema {
                SchemaUri = submission.ServiceType.Identifier,
                MimeType = MimeType.ApplicationJson,
            }
        };
        var contentStructure = new ContentStructure {
            Data = data,
            Attachments = submission.Attachments.Select(a =>
                new FitConnect.Models.Api.Metadata.Attachment {
                    Description = a.Description,
                    AttachmentId = a.Id,
                    MimeType = a.MimeType,
                    Filename = a.Filename,
                    Hash = new AttachmentHash() {
                        Type = HashType.Sha512,
                        Content = a.Hash
                    }
                }).ToList()
        };
        return JsonConvert.SerializeObject(contentStructure);
    }

    /// <summary>
    /// Retrieve the events for the submission
    /// </summary>
    /// <param name="caseId"></param>
    /// <returns></returns>
    /// <exception cref="NotImplementedException"></exception>
    public async Task<List<SecurityEventToken>> GetStatusForSubmission(string caseId) {
        if (!IsAuthenticated) {
            _oAuthToken = await AuthenticateAsync(Owner.ClientId, Owner.ClientSecret, null);
            SubmissionService.Token = _oAuthToken.AccessToken;
        }

        var events = SubmissionService.GetStatusForSubmissionAsync(caseId).Result;
        return events.Select(e => new SecurityEventToken(e)).ToList();
    }


    public CreateSubmissionDto AddSubmission(Submission submission) {
        return new CreateSubmissionDto {
            DestinationId = submission.DestinationId,
            ServiceType = (ServiceTypeDto)submission.ServiceType,
            AnnouncedAttachments = submission.Attachments.Select(a => a.Id).ToList()
        };
    }

    /// <summary>
    ///     Uploading the encrypted data to the server
    /// </summary>
    /// <param name="submissionId">Submissions ID</param>
    /// <param name="encryptedAttachments">Encrypted attachments with id and content</param>
    /// <exception cref="NotImplementedException"></exception>
    public async Task<bool> UploadAttachmentsAsync(string submissionId,
        Dictionary<string, string> encryptedAttachments) {
        // TODO: Upload encrypted attachments to server
        try {
            foreach (var (id, content) in encryptedAttachments) {
                Logger?.LogInformation("Uploading attachment {id}, content {content}", id, content);
                SubmissionService.UploadAttachment(submissionId, id, content);
            }

            return true;
        }
        catch (Exception e) {
            Logger?.LogError("Error Uploading attachment {message}", e.Message);
            throw;
        }
    }

    public async Task SubmitSubmissionAsync(Submission submission) {
        throw new NotImplementedException();
    }

    public Task<string> GetDestinationIdAsync(string leiaKey, string? ags, string? ars,
        string? areaId) {
        return RouteService.GetDestinationIdAsync(leiaKey, ags, ars, areaId);
    }


    /// <summary>
    ///     Encrypt Metadata with public key
    /// </summary>
    /// <param name="publicKey"></param>
    /// <param name="metadata"></param>
    /// <returns></returns>
    /// <exception cref="NotImplementedException"></exception>
    public string Encrypt(string publicKey, Metadata metadata) {
        throw new NotImplementedException();
    }

    /// <summary>
    ///     Encrypt data with public key (Fachdaten)
    /// </summary>
    /// <param name="publicKey"></param>
    /// <param name="data"></param>
    /// <returns></returns>
    /// <exception cref="NotImplementedException"></exception>
    public string Encrypt(string publicKey, Data data) {
        throw new NotImplementedException();
    }

    /// <summary>
    ///     Uploading the encrypted data and metadata to the server
    /// </summary>
    /// <param name="encryptedData"></param>
    /// <param name="encryptedMetadata"></param>
    /// <exception cref="NotImplementedException"></exception>
    public async Task<bool> SubmitAsync(string encryptedData, string encryptedMetadata) {
        throw new NotImplementedException();
    }

    private string Encrypt(object data, string publicKey) {
        var encryption = new FitEncryption(Logger) { PublicKeyEncryption = publicKey };
        return encryption.Encrypt(JsonConvert.SerializeObject(data));
    }

    // protected SubmitSubmissionDto CreateSubmitSubmissionDto(Submission submission) {
    //     var encryptData = Encrypt(PublicKey, submission.Data);
    //     var encryptMetadata = Encrypt(PublicKey, submission.Metadata);
    //
    //
    //     return new SubmitSubmissionDto {
    //         EncryptedData = encryptData,
    //         EncryptedMetadata = encryptMetadata
    //     };
    // }
}
